/*
 *  Copyright (C) 2002 Tomasz Kojm <zolw@konarski.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef CLAMUKO

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <pthread.h>
#include <clamav.h>

#include "server.h"
#include "others.h"
#include "cfgfile.h"
#include "dazukoio.h"
#include "clamuko.h"
#include "defaults.h"

struct access_t acc;

void clamuko_exit(int sig)
{

    logg("*Clamuko: clamuko_exit(), signal %d\n", sig);

    if(clamuko_scanning) {
	logg("*Clamuko: stopped while scanning %s\n", acc.filename);
	acc.deny = 0;
	dazukoReturnAccess(&acc); /* is it needed ? */
    }

    dazukoUnregister();
    clamuko_running = 0;
    logg("Clamuko stopped (exit).\n");
}

void *clamukoth(void *arg)
{
	struct thrarg *tharg = (struct thrarg *) arg;
	sigset_t sigset;
	char *virname;
        struct sigaction act;
	unsigned long mask = 0;
	const struct cfgstruct *pt;
	short int scan;
	int sizelimit = 0, options = 0;
	int maxwait = CL_DEFAULT_MAXWHILEWAIT;
	struct stat sb;


    clamuko_running = 1;

    /* ignore all signals except SIGUSR1 */
    sigfillset(&sigset);
    sigdelset(&sigset, SIGUSR1);
    pthread_sigmask(SIG_SETMASK, &sigset, NULL);
    act.sa_handler = clamuko_exit;
    sigfillset(&(act.sa_mask));
    sigaction(SIGUSR1, &act, NULL);

#ifdef C_LINUX
    logg("*Clamuko: Started in process %d\n", getpid());
#endif

    /* register */
    if(dazukoRegister()) {
	logg("!Clamuko: Can't register with Dazuko\n");
	clamuko_running = 0;
	return NULL;
    } else
	logg("Clamuko: Correctly registered with Dazuko.\n");

    /* access mask */
    if(cfgopt(tharg->copt, "ClamukoScanOnOpen")) {
	logg("Clamuko: Scan-on-open mode activated.\n");
	mask |= ON_OPEN;
    }
    if(cfgopt(tharg->copt, "ClamukoScanOnClose")) {
	logg("Clamuko: Scan-on-close mode activated.\n");
	mask |= ON_CLOSE;
    }
    if(cfgopt(tharg->copt, "ClamukoScanOnExec")) {
	logg("Clamuko: Scan-on-exec mode activated.\n");
	mask |= ON_EXEC;
    }

    if(!mask) {
	logg("!Access mask is not configured properly.\n");
	clamuko_running = 0;
	return NULL;
    }

    if(dazukoSetAccessMask(mask)) {
	logg("!Clamuko: Can't set access mask in Dazuko.\n");
	clamuko_running = 0;
	return NULL;
    }


    if((pt = cfgopt(tharg->copt, "ClamukoIncludePath"))) {
	while(pt) {
	    if((dazukoAddIncludePath(pt->strarg))) {
		logg("!Clamuko: Dazuko -> Can't include path %s\n", pt->strarg);
		clamuko_running = 0;
		return NULL;
	    } else
		logg("Clamuko: Included path %s\n", pt->strarg);

	    pt = (struct cfgstruct *) pt->nextarg;
	}
    } else {
	logg("!Clamuko: please include at least one path.\n");
	clamuko_running = 0;
	return NULL;
    }

    if((pt = cfgopt(tharg->copt, "ClamukoExcludePath"))) {
	while(pt) {
	    if((dazukoAddExcludePath(pt->strarg))) {
		logg("!Clamuko: Dazuko -> Can't exclude path %s\n", pt->strarg);
		clamuko_running = 0;
		return NULL;
	    } else
		logg("Clamuko: Excluded path %s\n", pt->strarg);

	    pt = (struct cfgstruct *) pt->nextarg;
	}
    }

    if(cfgopt(tharg->copt, "ClamukoScanArchive")) {
	options |= CL_ARCHIVE;
	logg("Clamuko: Archive support enabled.\n");
    } else {
	logg("Clamuko: Archive support disabled.\n");
    }

    if((pt = cfgopt(tharg->copt, "ClamukoMaxFileSize"))) {
	sizelimit = pt->numarg;
    } else
	sizelimit = CL_DEFAULT_CLAMUKOMAXFILESIZE;

    if(sizelimit)
	logg("Clamuko: Max file size limited to %d bytes.\n", sizelimit);
    else
	logg("Clamuko: File size limit disabled.\n");

    while(1) {

	/* wait while reloading the database */
	if(reload) {
	    logg("*Clamuko: Waiting (database reloading)\n");
	    clamuko_reload = 1;
	    maxwait = CL_DEFAULT_MAXWHILEWAIT;
	    while(reload && maxwait--)
		sleep(1);

	    if(!maxwait && reload) {
		logg("!Clamuko: Database reloading failed. Forcing quit...\n");
		logg("Clamuko stopped.\n");
		dazukoUnregister();
		kill(progpid, SIGTERM);
		clamuko_running = 0;
		clamuko_scanning = 0;
		clamuko_reload = 0;
		return NULL;
	    }

	    clamuko_reload = 0;
	}

	if(dazukoGetAccess(&acc) == 0) {
	    /* wait while reloading the database */
	    if(reload) {
		logg("*Clamuko: Waiting (database reloading, after dazukoGetAccess())\n");
		clamuko_reload = 1;
		maxwait = CL_DEFAULT_MAXWHILEWAIT;
		while(reload && maxwait--)
		    sleep(1);

		if(!maxwait && reload) {
		    logg("!Clamuko: Database reloading failed. Forcing quit...\n");
		    kill(progpid, SIGTERM);
		    acc.deny = 0;
		    dazukoReturnAccess(&acc);
		    logg("Clamuko stopped.\n");
		    dazukoUnregister();
		    clamuko_running = 0;
		    clamuko_scanning = 0;
		    clamuko_reload = 0;
		    return NULL;
		}
	    }

	    clamuko_scanning = 1;
	    scan = 1;

	    if(sizelimit) {
		stat(acc.filename, &sb);
		if(sb.st_size > sizelimit) {
		    scan = 0;
		    logg("*Clamuko: %s skipped (too big)\n", acc.filename);
		}
	    }

	    if(scan && cl_scanfile(acc.filename, &virname, NULL, tharg->root, tharg->limits, options) == CL_VIRUS) {
		logg("Clamuko: %s: %s FOUND\n", acc.filename, virname);
		acc.deny = 1;
	    } else
		acc.deny = 0;

	    if(dazukoReturnAccess(&acc)) {
		logg("!Can't return access to Dazuko.\n");
		logg("Clamuko stopped.\n");
		dazukoUnregister();
		clamuko_running = 0;
		clamuko_scanning = 0;
		return NULL;
	    }

	    clamuko_scanning = 0;
	}
    }

    /* can't be ;) */
    clamuko_running = 0;
    return NULL;
}

#endif
